@@ -2,15 +2,7 @@
 #
 # Copyright (c) 2005 Brenden Conte <conteb@cpan.org>, All Rights Reserved
 # Version 0.01
-# 2005-09-25 - Started work
-# 2005-12-08 - Module namespace approved by CPAN
-# 2006-02/06 - 0.01 released
 #
-# TODO
-#  - the objects 'server' hash does not reuse connections if the hostname is 
-#    slightly different.  Perhaps index on IP instead.
-#
-
 
 package IMAP::Client;
 
@@ -32,12 +24,16 @@ $|=1;
 our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
 
 @ISA = qw( Exporter );
-$VERSION = "0.03";
+$VERSION = "0.11";
 @EXPORT = qw ();
 @EXPORT_OK = qw();
 %EXPORT_TAGS = ();
 
-
+# Create Class variables
+my %Instances;
+my @SERVER_RESPONSES = ('exists', 'recent'); # constant value
+my $server_response_callback = undef;
+my $ID;
 
 =pod
 
@@ -55,7 +51,7 @@ $VERSION = "0.03";
     }
 (or)
     my $imap = new IMAP::Client();
-    imap->connect(PeerAddr => $server,
+    $imap->connect(PeerAddr => $server,
 		  ConnectMethod => 'SSL STARTTLS PLAIN',
 		  )
     or die "Unable to connect to [$server]: ".$imap->error();
@@ -64,6 +60,14 @@ $VERSION = "0.03";
     $imap->errorstyle('STACK');
     $imap->debuglevel(1);
     $imap->capability_checking(1);
+    
+	sub showstats ($) {
+    	my $resp = shift;
+	    foreach my $attr (keys %{$resp}) {
+			print "$attr: $resp->{$attr}\n";
+	    }
+	}
+	$imap->register_mailbox_update(\&showstats);
 
     $imap->authenticate($user,$pass)
         or die "Unable to authenticate as $user ".$imap->error()."\n";
@@ -77,6 +81,12 @@ $VERSION = "0.03";
 
  FIXME: more examples here
 
+=head1 IMPORTANT! READ THIS FIRST IF YOU ARE UPGRADING FROM PRE-0.10 TO 0.10 OR ABOVE!
+
+As of IMAP::Client 0.10, the "_active_server" mechanism has been removed, replaced instead by a class-wide monitoring of objects.  This means that if you have any code that utilizes the active_server functionality (using more than one connection in an instance of IMAP::Client), you will need to change your code to create seperate instances for each connection.
+
+Unfortunately, backward compatibility could not be maintained with this change.  However since tracking is now behind-the-scenes, this style should be the final one.
+
 =head1 DESCRIPTION
 
 This module was created as a low-level inteface to any IMAP server.  It was built to be a 'clear box' solution to working with an IMAP environment.  The idea is that anything an IMAP client should be able to do, and any information available via the IMAP specs, should be available to a client interface and user.  This way, the full strength of the IMAP protocol and data can be utilized, ideally in the most network-efficient mannger possible, rather than being contrained only to a subset of commands or data-limited responses.  If the server says it, the client should be able to see it.
@@ -87,7 +97,7 @@ This module also tries to follow the various RFCs for IMAPrev1 communications ve
     
 =over 4
 
-=item * RFC 3501 - INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1 
+=item * RFC 3501 - INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1 (partial)
 
 =item * RFC 2086 - IMAP4 ACL extension (0.01)
 
@@ -105,7 +115,7 @@ This module also tries to follow the various RFCs for IMAPrev1 communications ve
 
 =item * RFC 2359 - IMAP4 UIDPLUS extension (Partial in 0.01 - UID EXPUNGE check ok, need COPYUID and APPENDUID support)
 
-=item * RFC 2971 - IMAP4 ID extension (0.01)use Net::IMAP::Advanced;
+=item * RFC 2971 - IMAP4 ID extension (0.01)
 
 =item * RFC 3348 - IMAP4 Child Mailbox Extention (Not directly supported yet)
 
@@ -132,6 +142,21 @@ These are the lowest-level functions for use by the program.  These offer the mo
 =cut 
 
 ########## Internal, Undocumented Support functions ##########
+# First things first...
+sub dprint ($$$) {
+	my ($self, $debugbit,$string) = @_;
+	
+	# Error checking - make sure the debugbit is only 1 bit
+	unless ($debugbit && !($debugbit & ($debugbit - 1))) {
+		warn "Debugbit passed into dprint is not a single bit! (String = [$string])\n";
+	}
+	
+	if ($self->{DEBUG} & $debugbit) {
+		print STDERR $string;
+	}
+};
+
+
 # RFC3501 specifies a valid tag contains any char other than...
 my $RESP_SPECIALS = '\]';
 my $ATOM_SPECIALS = '\(\)\{ ';
@@ -151,11 +176,11 @@ sub untagged_ok_response(@) {
     return (($_[$#_] =~ /^\*\s*OK\s+/) ? 1 : 0);
 }
 sub failure_response(@) {
-    return (($_[$#_] =~ /^($VALID_TAG|\*)+?\s*(BAD|NO)\s+/) ? 1 : 0);
+    return (($_[$#_] =~ /^(?:$VALID_TAG|\*)+?\s*(BAD|NO)\s+/) ? 1 : 0);
 }
 
 sub is_sequence_set($) {
-    return (($_[0] =~ /^(\d+|\d+\:\d+)(\,\d+|\d+\:\d+)*$/) ? 1 : 0);
+    return (($_[0] =~ /^(?:(?:\d+|\d+\:\d+)(?:\,\d+|\d+\:\d+)*|\d+:\*|\*)$/) ? 1 : 0);
 }
 sub sequencify (@){ # preserves ordering over compression
     my $string;
@@ -174,6 +199,7 @@ sub sequencify (@){ # preserves ordering over compression
     $string .= ($start == $end) ? "$start" : "$start:$end"; # last entry
     return($string);
 }
+
 sub throw_error($$) {
     my ($self,$error) = @_;
     $error =~ s/^(.*?)\s*\r?\n?$/$1/;
@@ -188,7 +214,7 @@ sub throw_error($$) {
 		$self->{error_read} = 0;
 		return (undef);
     } elsif ($self->{onfail} eq 'abort') {
-		print STDERR $newerror;
+		print STDERR $newerror."\n";
 		exit(-1);
     } else {
 		print STDERR "INTERNAL ERROR: Unknown failure handler string [",$self->{onfail},"], aborting...\n";
@@ -198,7 +224,7 @@ sub throw_error($$) {
 sub parse_select_examine(@) {
 	return() unless $_[0];
     my %ret;
-    my ($_t, $_v);
+    my ($_t, $_v); # t is the TITLE (or label/tag), and v is the VALUE of t.
     foreach my $line (@_) {
 		if (ok_response($line)) { # done
 		    my ($perm) = $line =~ /\[([\w-]+)\]/;
@@ -210,6 +236,8 @@ sub parse_select_examine(@) {
 		    $ret{$_t} = $_v;
 		} elsif (($_v, $_t) = $line =~ /(\d+)\s*(\w+)/) { # num-title: # TITLE
 		    $ret{$_t} = $_v;
+		} elsif (($_t, $_v) = $line =~ /NO\s+\[(.*?)\]\s+(.*)$/) { # NO [TITLE] VALUE (usually for ALERTs)
+			$ret{$_t} = $_v;
 		} else {
 		    warn "Unknown tagless response(): $line\n";
 		}
@@ -221,7 +249,7 @@ sub parse_list_lsub(@) {
     my @list;
     foreach my $line (@result) {
 		next if (ok_response($line));
-		my ($flags,$reference,$mailbox) = $line =~ /^\*\s+LIST\s+\((.*?)\)\s+\"(.*)\"\s+\"(.*)\"\r\n$/;
+		my ($flags,$reference,$mailbox) = $line =~ /^\*\s+LIST\s+\((.*?)\)\s+\"(.*?)\"\s+\"?(.*?)\"?\r\n$/;
 		my %hash = (FLAGS => $flags, REFERENCE => $reference, MAILBOX => $mailbox);
 		push @list, \%hash;
     }
@@ -298,11 +326,11 @@ sub parse_envelope($) {
     $ret{'FROM'} = $_t if ($_t = address($from));
     $ret{'SENDER'} = $_t if ($_t = address($sender));
     $ret{'REPLYTO'} = $_t if ($_t = address($replyto));
-    $ret{'TO'} = $_t if ($_t = address($from));
+    $ret{'TO'} = $_t if ($_t = address($to));
     $ret{'CC'} = $_t if ($_t = address($cc));
     $ret{'BCC'} = $_t if ($_t = address($bcc));
     $ret{'INREPLYTO'} = $_t if ($_t = address($inreplyto));
-    $ret{'MESSAGEID'} = $_t if ($_t = address($messageid));
+    $ret{'MESSAGEID'} = $_t if ($_t = debracket(dequote($messageid)));
     return(\%ret);
 }
 # recursive function for building body hash
@@ -425,35 +453,53 @@ sub parse_fetch ($@) {
     my ($self,@resp) = @_;
     return(undef) unless ($resp[0]);
 
-    my $fetchset = join('',@resp);
+    # Load up hash with fetch results (one entry per * tag FETCH)
+    my %fetchsets;
+    my $msgid = -1;
+    foreach my $line (@resp) {
+		if ($line =~ /^\* (\d+) FETCH \(/gs) {
+			$msgid = $1;
+			$fetchsets{$msgid} .= $line;
+		} elsif (ok_response($line)) {
+			$msgid = -1; # We Found the OK - This should be the end of the command
+		} else {
+			$self->throw_error("INTERNAL ERROR: No msgid set, but still trying to build fetchsets\n") if ($msgid < 0);
+			$fetchsets{$msgid} .= $line;
+		}
+    }
+
+    $self->dprint(0x2, "-- parse_fetch: Fetch count: ".scalar(keys %fetchsets)."\n");
     my %results;
 
+	#$self->dprint(0x2, "-- parse_fetch: evaluating [$fetchset]\n");
+
     # find FETCH lines and process results
-    while (my ($msgid) = ($fetchset =~ /^\* (\d+) FETCH \(/gs)) {
+	foreach my $msgid (keys %fetchsets) {
+		$self->dprint(0x2, "-- parse_fetch: Parsing FETCH response [$fetchsets{$msgid}]\n");
 		my %ret;
 		my $len = length($msgid)+10; # get length of just-extracted line
-		$fetchset = substr($fetchset, $len); # remove it from the remainder of the FETCH response
+		$fetchsets{$msgid} = substr($fetchsets{$msgid}, $len); # remove fetch line from the remainder of the FETCH response
+		$self->dprint(0x2, "-- parse_fetch: FETCH response after FETCH line removed: [$fetchsets{$msgid}]\n");
 		# Break into hash (unfortunately, we can't do it inline regexp, since thre is a 32765 char limit on {min,max}. grr)
 		my %result_entries;
-		while (!($fetchset =~ /^(?:\* (\d+) FETCH \(|\r\n\S+ OK)/)) {
-		    my ($key, $length) = ($fetchset =~ /^(BODY|BODYSTRUCTURE|ENVELOPE|FLAGS|INTERNALDATE|UID|RFC822.SIZE|BODY\[.*?\](?:\<\d+\>)?|RFC822(?:\.TEXT|\.HEADER)?) (?:\{(\d+)\}\r?\n?)?/gis) or return($self->throw_error("INTERNAL ERROR: unable to find keys in [".substr($fetchset,0,20)));
-		    $fetchset = substr($fetchset,(length($key)+1)+(($length) ? length($length)+4 : 0)); # trim newly found entries
+		while ($fetchsets{$msgid}) {
+		    my ($key, $length) = ($fetchsets{$msgid} =~ /^(BODY|BODYSTRUCTURE|ENVELOPE|FLAGS|INTERNALDATE|UID|RFC822.SIZE|BODY\[.*?\](?:\<\d+\>)?|RFC822(?:\.TEXT|\.HEADER)?) (?:\{(\d+)\}\r?\n?)?/gis) or return($self->throw_error("INTERNAL ERROR: unable to find keys in [".substr($fetchsets{$msgid},0,20)."...]"));
+		    $fetchsets{$msgid} = substr($fetchsets{$msgid},(length($key)+1)+(($length) ? length($length)+4 : 0)); # trim newly found entries
 		    if ($length) {
 				$result_entries{$key}{'length'} = $length;
-				$result_entries{$key}{'value'} = substr($fetchset,0,$length); # Save length of value
+				$result_entries{$key}{'value'} = substr($fetchsets{$msgid},0,$length); # Save length of value
+
 				if (length($result_entries{$key}{'value'}) < $length) {
-				    return($self->throw_error("INTERNAL ERROR: unable to get [$length] length of fetchset [".length($fetchset)." available]"));
+				    return($self->throw_error("INTERNAL ERROR: unable to get [$length] length of fetchset [".length($fetchsets{$msgid})." available]"));
 				}
-			$fetchset = substr($fetchset,$length); # trim length of message
+			$fetchsets{$msgid} = substr($fetchsets{$msgid},$length); # trim length of message
 		    } else { # no length, just a value
-				($result_entries{$key}{'value'}) = ($fetchset =~ /^($parens|$nstring|$number)/gis) 
-				    or return($self->throw_error("INTERNAL ERROR: No value in [".substr($fetchset,0,20)."]"));
-				$fetchset = substr($fetchset,length($result_entries{$key}{'value'}));
+				($result_entries{$key}{'value'}) = ($fetchsets{$msgid} =~ /^($parens|$nstring|$number)/gis) 
+				    or return($self->throw_error("INTERNAL ERROR: No value in [".substr($fetchsets{$msgid},0,20)."]"));
+				$fetchsets{$msgid} = substr($fetchsets{$msgid},length($result_entries{$key}{'value'}));
 		    }
-		    # trim end-of-command space
-		    $fetchset = substr($fetchset,1) if ($fetchset);
+		    $fetchsets{$msgid} =~ s/^[\)\r\n\s]*// if ($fetchsets{$msgid}); # Remove all 'unessesary trailers' (varies depending on end-of-command, end-of-line, etc)
 		}
-		$fetchset = substr($fetchset,2) if ($fetchset =~ /"\r\n"/);
 
 
 		 # Ok, we have our entries for this msgid - store them
@@ -563,13 +609,13 @@ Sends the string argument provided to the server.  A tag is automatically prepen
     
 sub imap_send ($$) {
     my ($self,$string) = @_;
-#    return($self->throw_error("No servers defined for [$string]")) unless $self->{'_active_server'};
-    my $server = $self->{'server'}->{$self->{'_active_server'}};
+#    return($self->throw_error("No servers defined for [$string]")) unless $self->{'server'};
+    my $server = $self->{'server'};
 
     my $tag = sprintf("%04i",++$self->{tag});
     chomp($string);
 
-    print ">> $tag $string\r\n" if ($self->{DEBUG});
+    $self->dprint(0x1, ">> $tag $string\r\n");
     print $server "$tag $string\r\n";
 
     $self->{tag} = $tag;
@@ -585,16 +631,16 @@ Send the string argument provided to the server B<without a tag>.  This is neede
 
 sub imap_send_tagless ($$) {
     my ($self,$string) = @_;
-#    return($self->throw_error("No servers defined for [$string]")) unless $self->{'_active_server'};
-    my $server = $self->{'server'}->{$self->{'_active_server'}};
+#    return($self->throw_error("No servers defined for [$string]")) unless $self->{'server'};
+    my $server = $self->{'server'};
 
-    print ">> $string\r\n" if ($self->{DEBUG});
+    $self->dprint(0x1, ">> $string\r\n");
     print $server "$string\r\n";
 }
 
 =pod
 
-=item B<imap_receive()>
+=item B<imap_receive()> 
 
 Accept responses from the server until the previous tag is encountered, at which point return the entire response.  CAUTION: can cause your program to hang if a tagged response isn't given.  For example, if the command expects more input and issues a '+' response, and waits for input, this function will never return. 
 
@@ -603,11 +649,11 @@ Accept responses from the server until the previous tag is encountered, at which
 # FIXME: A timeout option would be nice...
 sub imap_receive($) {
     my ($self) = @_;
-#    return($self->throw_error("No servers defined for receiving")) unless $self->{'_active_server'};
+#    return($self->throw_error("No servers defined for receiving")) unless $self->{'server'};
     my (@r, $_t);
     do {
-		$_t = $self->{'server'}->{$self->{'_active_server'}}->getline; 
-		print "<< $_t" if ($self->{DEBUG});
+		$_t = $self->{'server'}->getline; 
+		$self->dprint(0x1, "<< $_t");
 		push (@r, $_t);
     } until ($_t =~ /^$self->{tag}/);
 	if ($#r < 1) {
@@ -627,10 +673,9 @@ Accept a line of response, regardless if a tag is provided.  Misuse of this func
 
 sub imap_receive_tagless($) {
     my ($self) = @_;
-#    return($self->throw_error("No servers defined for receiving")) unless ($self->{'_active_server'});
-    my ($_t);
-    $_t = $self->{'server'}->{$self->{'_active_server'}}->getline; 
-    print "<< $_t" if ($_t && $self->{DEBUG});
+    my $_t;
+    $_t = $self->{'server'}->getline; 
+    $self->dprint(0x1, "<< $_t") if ($_t);
     return ($_t);
 }
 
@@ -659,23 +704,26 @@ Creates a new IMAP object.  You can optionally specify the server to connect to,
 =cut
 
 sub new ($){
-    my (undef) = shift; #ignore first arg
-    my $self = {server => {},
-		_active_server => undef,
-		tag => '*',
-		error => '',
-		error_read => 1,
-		DEBUG => 0,
-		onfail => 'error',
-		errorstyle => 'stack',
-		capability => '',
-		capabilities => {},
-		capability_checking => 1,
-		user => '',
-		auth => '',
-	    };
-    bless $self,"IMAP::Client";
-
+    my $proto = shift; #ignore first arg
+    my $self = {};
+    bless $self = {
+		'ID' => $ID++,
+    	'name' => '', # the DNS/IP address to match
+    	'server' => '', # Actual socket
+		'tag' => '*',
+		'error' => '',
+		'error_read' => 1,
+		'DEBUG' => 0,
+		'onfail' => 'error',
+		'errorstyle' => 'stack',
+		'capability' => '',
+		'capabilities' => {},
+		'capability_checking' => 1,
+		'user' => '',
+		'auth' => '',
+	    }, $proto;
+	
+	$Instances{$self->{'ID'}} = \$self;
     # If a server was supplied, try to connect to it
     if (my $server = shift) {
 		if ($self->connect(PeerAddr => $server)) {
@@ -689,23 +737,57 @@ sub new ($){
 		return($self);
     }
 }
-
+sub DESTROY { # Undocumented: Not to be directly called
+	my $self = shift;
+	if ($Instances{$self->{'ID'}}) {
+		$self->disconnect(); # FIXME: Probably not nessesary
+		delete $Instances{$self->{'ID'}};
+		return;
+	}
+	
+	# If we reach here, we didn't find ourself, which is a seirous problem
+	warn "ERROR: Could not find self in Instances upon DESTROY - Something is seriously wrong!\n";
+	foreach my $key (keys %Instances) {
+		warn "PANIC DUMP: Instance name: ".(($Instances{$key}->{'name'}) ?  $Instances{$key}->{'name'} : "(none)")."\n";
+	}
+	die "DUMP COMPLETE: Aborting...\n\n";
+}
 
 ##### Object manipulation functions
 =pod
 
 =item B<debuglevel($level)>
 
-Set the debug level.  Valid values are 0 (for no debugging) or 1 (for a communications dump).
+Set the debug level.  Debug levels are set on a bitmask, and all debug output is to STDERR.  Valid bits are as follows:
+
+=over 4
+
+=item B<0x1 - Communications dump>
+
+=over 2
+
+This will output all IMAP communications, with >> showing the sent data, and << showing the received data.
+
+=back
+
+=item B<0x2 - Fetch-parsing dump>
+
+=over 2
+
+This dumps a *lot* of data about the parsing of a fetch statement.  
+
+=back
+
+=back
 
 =cut
 
 sub debuglevel($$) {
     my ($self,$level) = @_;
-    if (($level >= 0) && ($level <= 1)) {
+    if ($level >= 0)  {
 		$self->{DEBUG} = $level;
     } else {
-		return($self->throw_error("Invalid value [$level] for debuglevel: valid values are 0 or 1"));
+		return($self->throw_error("Invalid value [$level] for debuglevel: valid values are 0 and above"));
     }
 }
 
@@ -789,19 +871,53 @@ sub _imap_command ($$@) {
     my ($self,$command, @argset) = @_;
     my @fullresp;
     my $i=0;
-    return($self->throw_error("No servers defined for [$command][".join('][',@argset)."]")) unless $self->{'_active_server'};
+    return($self->throw_error("No servers defined for [$command][".join('][',@argset)."]")) unless $self->{'server'};
 
     $self->imap_send(($argset[0]) ? "$command $argset[0]" : $command);
     
+    my ($datasize,$datasizemax) = (0,0);
     while (1) {
 		my $resp = $self->imap_receive_tagless();
 		if ($resp) {
 		    push(@fullresp,$resp);
+
+			# First, check to see if we are in the middle of a size-based continuation response
+			if ($datasizemax > 0) {
+				$datasize += length($resp) + 1;
+				# >= because we could get termination chars (like a ')n', for instance) in the last line.  
+				if ($datasize >= $datasizemax) { 
+					($datasize,$datasizemax) = (0,0);
+				}
+				next; # next line
+			}
+
+			# First, check if this is a multi-line response
+		    if (untagged_response($resp)) { 
+				if ($resp =~ /\{(\d+)\}\s*$/) {
+					$datasizemax = $1; 
+				}
+				next;
+		    } 
 		    
 		    if (ok_response($resp)) {
 				if ($i != $#argset) {
 				    print STDERR "WARNING: Only ", $i+1 ," arguments of ", $#argset+1 ," used before successful response in [$command] command\n";
 				}
+				# Check for 'server responses' as defined in RFC3501, Section 7.
+				if ($server_response_callback) {
+					my %actual_responses;
+					foreach my $line (@fullresp) {
+						if (my ($_v, $_t) = $line =~ /(\d+)\s*(\w+)/) { # num-title: # TITLE
+							foreach my $attr (@SERVER_RESPONSES) { # We only match pre-approved values (FIXME: Is this what we want?)
+								if (lc($_t) eq lc($attr)) {	
+									$actual_responses{$attr} = $_v;
+								}
+							}
+						}
+					}
+					$server_response_callback->(\%actual_responses); # Run the callback with the values
+				}
+			    # Return the results
 				return(@fullresp);
 		    } elsif (continue_response($resp)) {
 				if ($i < $#argset) {
@@ -809,10 +925,10 @@ sub _imap_command ($$@) {
 				} else {
 				    return($self->throw_error("$command failed: Server wanted more continuations than the ".$#argset." provided"));
 				}
-		    } elsif (untagged_response($resp)) {
-				# do nothing, keep reading
 		    } elsif (failure_response($resp)) {
 				return($self->throw_error("$command failed: @fullresp"));
+		    } elsif (untagged_response($resp)) { 
+				# This is just to avoid having untagged responses flagged as 'unrecognised'
 		    } else {
 				# unrecognized response - put in any times its ok for this to happen in the unless statement.
 				unless ((lc($command) eq 'fetch') || (lc($command) eq 'uid fetch')) {
@@ -908,10 +1024,7 @@ sub connect($%) {
 		# but immediately dump the connection if something isn't supported (i.e. Exchange and
 		# connecting to Non-ssl when SSL is required by the server)
 	
-		# UNCLEAN: the server has to be preset for communications to work
-		my $prev_active_server = $self->{'_active_server'};
-		$self->{'server'}->{$args{'PeerAddr'}} = $server;
-		$self->{'_active_server'} = $args{'PeerAddr'};
+		$self->{'server'} = $server;
 		@resp = $self->imap_receive_tagless(); # collect welcome
 		if ($resp[0] && untagged_ok_response(@resp) && ok_response($self->noop())) {
 		    # Post-processing
@@ -929,17 +1042,13 @@ sub connect($%) {
 		} else {
 		    $errorstr .= "$method attempt: Connection dropped upon connect\n";
 		}
-		# restore old settings (if we're successful, we'll *properly* set this up below)
-		$self->{'_active_server'} = $prev_active_server;
-		delete $self->{'server'}->{$args{'PeerAddr'}}; #probably unessesary
     }
 	    
     if (!$connected) {
 		chop($errorstr); # clip the tailing newline: we print errors without them
 		return ($self->throw_error($errorstr));
     } else {
-		$self->{'server'}->{$args{'PeerAddr'}} = $server;
-		$self->{'_active_server'} = $args{'PeerAddr'} unless ($self->{'server'});
+		$self->{'server'} = $server;
 		$self->error; # clear error logs
     }
     
@@ -950,33 +1059,42 @@ sub connect($%) {
 
 =item B<disconnect($server)>
 
-Disconnect from the specified server.  Must match the server name used to connect.
+Disconnect from the server.  This command can safely be used on an already-disconnected server.
 
 =cut
 
 sub disconnect($$) {
     my ($self,$server) = @_;
 
-    return($self->throw_error("Server [$server] not found"))
-		unless $self->{'server'}->{$server};
+	if ($self->{'server'}) { # If we're still connected to something...
+	    $self->{'server'}->close(); # close connection
+		undef $self->{'server'}; # remove server
+	    undef $self->{'name'}; # clear name
+	    undef $self->{'auth'}; # clear authentication info
+	    $self->{'tag'} = '*'; # Reset tag to * (for welcome message)
+	}
+}
 
-    if ($server eq $self->{'_active_server'}) {
-		# randomly choose a new one, if available
-		foreach my $key (keys %{$self->{'server'}}) {
-		    if ($key ne $server) {
-				$self->{'_active_server'} = $key;
-				last;
-		    }
-		}
-		# ensure we have a new one - if not, we are the last
-		if ($server eq $self->{'_active_server'}) {
-		    $self->{'_active_server'} = undef;
-		}
-    }
-    $self->{'server'}->{$server}->close(); # close connection
-    delete $self->{'server'}->{$server}; # remove server
+=pod 
 
-    return(1);
+=item B<register_mailbox_update(&subfunction)>
+
+This function sets up a callback function for when the IMAP server sends back a Server Response in accordance with RFC3501 Section 7.3, which stiuplates that while select()ed or examine()ing a mailbox, updates as to the mailbox content can be sent back after any command as tagless responses.
+
+The subfunction will be passed one hash argument.  The hash argument will contain keys that represent the data type (EXISTS, RECENT), and their respective values will be the values returned by the server.
+
+If no function is registered via this method, or this method is called with an 'undef' argument, no special action will be taken should these server responses be encountered by the library.
+
+=cut 
+
+sub register_mailbox_update($$) {
+	my ($self, $callback_func) = @_;
+
+	# Test to make sure we have a sub reference (or undef to clear the callback)
+	return($self->throw_error("Argument to register_mailbox_update is not a CODE reference\n"))
+		unless ((ref($callback_func) eq 'CODE') || (!defined $callback_func)); 
+	
+	$server_response_callback = $callback_func;
 }
 
 =pod 
@@ -1079,8 +1197,8 @@ sub starttls ($){
 		return($self->throw_error("STARTTLS not found in CAPABILITY"));
     }
     my @recv = $self->_imap_command("STARTTLS",undef);
-    print "<TLS negotiations>\n" if $self->{DEBUG}; # compensation for lack of tapping into dump
-    if (IO::Socket::SSL->start_SSL($self->{'server'}->{$self->{'_active_server'}})) {
+    $self->dprint(0x1, "<TLS negotiations>\n"); # compensation for lack of tapping into dump
+    if (IO::Socket::SSL->start_SSL($self->{'server'})) {
 		# per RFC 3501 - 6.2.1, we must re-establish the CAPABILITY of the server after STARTTLS
 		$self->{capability} = '';
 		@recv = $self->capability();
@@ -1159,16 +1277,19 @@ Open a mailbox in read-write mode so that messages in the mailbox can be accesse
 
 =item * OK [UIDNEXT <n>]
 
-=back
 
 If the server supports an earlier version of the protocol than IMAPv4, the only flags required are FLAGS, EXISTS, and RECENT.
 
 Finally, hash responses will have an 'OK' key that will contain the current permissional status, either 'READ-WRITE' or 'READ-ONLY', if returned by the server.  Returns an empty (undefined) hash on error.
 
+IMPORTANT!  You should always check to see if an ALERT was issued.  ALERTs should be relayed to the user if they exist!
+
 =cut
 
 sub select($$) {
     my ($self,$mailbox) = @_;
+    # Don't add quotes around mailbox if they already exist
+	$mailbox = "\"$mailbox\"" unless ($mailbox =~ /^\".*\"$/);
     return(parse_select_examine($self->_imap_command("SELECT", $mailbox)));
 }
 
@@ -1182,6 +1303,7 @@ Identical to select(), except the mailbox is opened READ-ONLY.  Returns an empty
 
 sub examine($$) {
     my ($self,$mailbox) = @_;
+	$mailbox = "\"$mailbox\"" unless ($mailbox =~ /^\".*\"$/);
     return(parse_select_examine($self->_imap_command("EXAMINE", $mailbox)));
 }
 
@@ -1204,11 +1326,15 @@ For example:
 
 =cut
 
-sub create($$) {
-    my ($self,$mailbox,$properties) = @_;
+sub create($@) {
+    my ($self,$mailbox,$properties,$server) = @_;
 
     # Create mailbox
-    return undef unless ($self->_imap_command("CREATE", $mailbox)); #err already thrown
+    if ($server) {
+    	return undef unless ($self->_imap_command("CREATE", "$mailbox $server")); #err already thrown
+    } else {
+	    return undef unless ($self->_imap_command("CREATE", $mailbox)); #err already thrown
+    }
     
     # set quota (if needed)
     if ($properties->{quota}) {
@@ -1377,6 +1503,8 @@ This 'CRLF' assurance is done for all commands - however its noted here because
 
 Unless overridden, append will check for the LITERAL+ capability() atom, and use non-synchronizing literals if supported - otherwise, it will use the standard IMAP dialog.
 
+Upon successful execution, the return of this function depends on the type of variable receiving the data.
+
 =cut
 
 sub append ($$$) {
@@ -1397,11 +1525,13 @@ sub append ($$$) {
 
     $flaglist = "()" unless $flaglist;
 
-    if ($self->check_capability('LITERAL+')) { #non-synchronizing literals support
-		return($self->_imap_command("APPEND","$mailbox $flaglist {$messagelen+}\r\n$message"));
-    } else {
-		return($self->_imap_command("APPEND","$mailbox $flaglist {$messagelen}",$message));
-    }
+	my @result = ($self->check_capability('LITERAL+')) ? #non-synchronizing literals support
+					$self->_imap_command("APPEND","$mailbox $flaglist {$messagelen+}\r\n$message") :
+					$self->_imap_command("APPEND","$mailbox $flaglist {$messagelen}",$message);
+					
+	if ($self->check_capability('UIDPLUS')) {
+		
+	}
 }
 
 # selected state
@@ -1686,8 +1816,11 @@ sub uidstore($$$$) {
 
 Identical to the expunge() command, except you can specify a set of messages to be expunged, rather than the entire mailbox, via a UID set.  This function ensures the existance of the UIDPLUS atom in the capability() command.
 
+Note: At this time, the function does not implement the reccomendation in RFC2359, which suggestests that clients use alternate methods in selectivly expunging messages on servers that do not support UIDPLUS.
+
 =cut
 
+#FIXME This needs to implement the UIDEXPUNGE suggested alternate methods.
 sub uidexpunge($$) { 
     my ($self,$sequence) = @_;
     return($self->throw_error("Invalid sequence string: $sequence")) unless (is_sequence_set($sequence));
